package com.emamaker.voxeltest.intervaltrees.utils;

public class SpaceFillingCurve {
    // http://and-what-happened.blogspot.com/2011/08/fast-2d-and-3d-hilbert-curves-and.html
    public static int[][][] MORTON_XYZ_ENCODE;
    public static short[][] MORTON_XYZ_DECODE;
    public static int[][][] HILBERT_XYZ_ENCODE;
    public static short[][] HILBERT_XYZ_DECODE;

    public static void initLUT() {
        MORTON_XYZ_ENCODE = new int[Config.CHUNK_SIZE][Config.CHUNK_SIZE][Config.CHUNK_SIZE];
        MORTON_XYZ_DECODE = new short[Config.CHUNK_3DCOORD_MAX_INDEX + 1][3];
        HILBERT_XYZ_ENCODE = new int[Config.CHUNK_SIZE][Config.CHUNK_SIZE][Config.CHUNK_SIZE];
        HILBERT_XYZ_DECODE = new short[Config.CHUNK_3DCOORD_MAX_INDEX + 1][3];

       int morton, hilbert;
        for (short i = 0; i < Config.CHUNK_SIZE; i++)
            for (short j = 0; j < Config.CHUNK_SIZE; j++)
                for (short k = 0; k < Config.CHUNK_SIZE; k++) {
                    morton = Morton_3D_Encode_10bit(i,j,k);
                    hilbert = MortonToHilbert3D(morton, Config.HILBERT_BITS);

                    MORTON_XYZ_ENCODE[i][j][k] = morton;
                    MORTON_XYZ_DECODE[morton][0] = i;
                    MORTON_XYZ_DECODE[morton][1] = j;
                    MORTON_XYZ_DECODE[morton][2] = k;

                    HILBERT_XYZ_ENCODE[i][j][k] = hilbert;
                    HILBERT_XYZ_DECODE[hilbert][0] = i;
                    HILBERT_XYZ_DECODE[hilbert][1] = j;
                    HILBERT_XYZ_DECODE[hilbert][2] = k;
                }
    }

    public static int Morton_3D_Encode_10bit(short index1, short index2, short index3) { // pack 3 10-bit indices into a 30-bit Morton code
        index1 &= 0x000003ff;
        index2 &= 0x000003ff;
        index3 &= 0x000003ff;
        index1 |= (index1 << 16);
        index2 |= (index2 << 16);
        index3 |= (index3 << 16);
        index1 &= 0x030000ff;
        index2 &= 0x030000ff;
        index3 &= 0x030000ff;
        index1 |= (index1 << 8);
        index2 |= (index2 << 8);
        index3 |= (index3 << 8);
        index1 &= 0x0300f00f;
        index2 &= 0x0300f00f;
        index3 &= 0x0300f00f;
        index1 |= (index1 << 4);
        index2 |= (index2 << 4);
        index3 |= (index3 << 4);
        index1 &= 0x030c30c3;
        index2 &= 0x030c30c3;
        index3 &= 0x030c30c3;
        index1 |= (index1 << 2);
        index2 |= (index2 << 2);
        index3 |= (index3 << 2);
        index1 &= 0x09249249;
        index2 &= 0x09249249;
        index3 &= 0x09249249;
        return (index1 | (index2 << 1) | (index3 << 2));
    }

    public static short[] Morton_3D_Decode_10bit(int morton) { // unpack 3 10-bit indices from a 30-bit Morton code
        int value1 = morton;
        int value2 = (value1 >> 1);
        int value3 = (value1 >> 2);
        value1 &= 0x09249249;
        value2 &= 0x09249249;
        value3 &= 0x09249249;
        value1 |= (value1 >> 2);
        value2 |= (value2 >> 2);
        value3 |= (value3 >> 2);
        value1 &= 0x030c30c3;
        value2 &= 0x030c30c3;
        value3 &= 0x030c30c3;
        value1 |= (value1 >> 4);
        value2 |= (value2 >> 4);
        value3 |= (value3 >> 4);
        value1 &= 0x0300f00f;
        value2 &= 0x0300f00f;
        value3 &= 0x0300f00f;
        value1 |= (value1 >> 8);
        value2 |= (value2 >> 8);
        value3 |= (value3 >> 8);
        value1 &= 0x030000ff;
        value2 &= 0x030000ff;
        value3 &= 0x030000ff;
        value1 |= (value1 >> 16);
        value2 |= (value2 >> 16);
        value3 |= (value3 >> 16);
        value1 &= 0x000003ff;
        value2 &= 0x000003ff;
        value3 &= 0x000003ff;

        return new short[]{(short) value1, (short) value2, (short) value3};
    }

    public static int MortonToHilbert3D(int morton, int bits) {
        int hilbert = morton;
        if (bits > 1) {
            int block = ((bits * 3) - 3);
            int hcode = ((hilbert >> block) & 7);
            int mcode, shift, signs;
            shift = signs = 0;
            while (block > 0) {
                block -= 3;
                hcode <<= 2;
                mcode = ((0x20212021 >> hcode) & 3);
                shift = ((0x48 >> (7 - shift - mcode)) & 3);
                signs = ((signs | (signs << 3)) >> mcode);
                signs = ((signs ^ (0x53560300 >> hcode)) & 7);
                mcode = ((hilbert >> block) & 7);
                hcode = mcode;
                hcode = (((hcode | (hcode << 3)) >> shift) & 7);
                hcode ^= signs;
                hilbert ^= ((mcode ^ hcode) << block);
            }
        }
        hilbert ^= ((hilbert >> 1) & 0x92492492);
        hilbert ^= ((hilbert & 0x92492492) >> 1);
        return (hilbert);
    }

    public static int HilbertToMorton3D(int hilbert, int bits) {
        int morton = hilbert;
        morton ^= ((morton & 0x92492492) >> 1);
        morton ^= ((morton >> 1) & 0x92492492);
        if (bits > 1) {
            int block = ((bits * 3) - 3);
            int hcode = ((morton >> block) & 7);
            int mcode, shift, signs;
            shift = signs = 0;
            while (block > 0) {
                block -= 3;
                hcode <<= 2;
                mcode = ((0x20212021 >> hcode) & 3);
                shift = ((0x48 >> (4 - shift + mcode)) & 3);
                signs = ((signs | (signs << 3)) >> mcode);
                signs = ((signs ^ (0x53560300 >> hcode)) & 7);
                hcode = ((morton >> block) & 7);
                mcode = hcode;
                mcode ^= signs;
                mcode = (((mcode | (mcode << 3)) >> shift) & 7);
                morton ^= ((hcode ^ mcode) << block);
            }
        }
        return (morton);
    }
    }