package com.emamaker.voxeltest.intervaltrees.world.chunk;

import com.emamaker.voxeltest.intervaltrees.data.IntervalMap;
import com.emamaker.voxeltest.intervaltrees.utils.Config;
import com.emamaker.voxeltest.intervaltrees.utils.Utils;
import com.emamaker.voxeltest.intervaltrees.world.blocks.Blocks;
import com.jme3.scene.Node;

public class Chunk {

    public int posx, posy, posz;
    public IntervalMap<Blocks> blocks = new IntervalMap<>();
    public Node chunkNode = new Node();
    // A bit field representing the state of the chunk, where each bit is an operation done on the chunk
    private byte state = 0;
    // Convenient access to the bit field, each state is the position in the byte (0 is LSB)
    public static final byte CHUNK_STATE_GENERATED = 1;
    public static final byte CHUNK_STATE_MESHED = 2;
    public static final byte CHUNK_STATE_LOADED = 4;
    public static final byte CHUNK_STATE_EMPTY = 64;

    public Chunk() {
        this(0,0,0);
    }

    public Chunk(int x_, int y_, int z_) {
        this.posx = x_;
        this.posy = y_;
        this.posz = z_;
        chunkNode.setLocalTranslation(posx, posy, posz);
        chunkNode.setLocalTranslation(chunkNode.getLocalTranslation().mult(Config.CHUNK_SIZE).mult(Config.VOXEL_SIZE));
        
        setBlocks(Blocks.AIR, 0, Config.CHUNK_3DCOORD_MAX_INDEX + 1);
        setState(CHUNK_STATE_EMPTY, true);
    }

    public void setState(byte nstate, boolean value){
       if(value) state |= nstate;
       else state &= ~nstate;
    }

    public boolean bgetState(byte nstate){
        return (state & nstate) != 0;
    }
    public int getState(byte nstate){
        return state & nstate;
    }
    public static int coord3DTo1D(int x, int y, int z) {
        return Utils.coord3DTo1D(x, y, z, Config.CHUNK_SIZE, Config.CHUNK_SIZE, Config.CHUNK_SIZE);
    }

    public static int[] coord1DTo3D(int idx) {
        return Utils.coord1DTo3D(idx, Config.CHUNK_SIZE, Config.CHUNK_SIZE, Config.CHUNK_SIZE);
    }

    /*
     * Set blocks. Interval to be intended in the same way as in interval maps
     */
    public void setBlocks(Blocks block, int intLow, int intHigh) {
        if(intHigh > intLow)
            this.blocks.insert( Math.max(0,intLow), Math.min(Config.CHUNK_3DCOORD_MAX_INDEX+1, intHigh),block);
    }

    public Blocks getBlock(int x, int y, int z) {
        return blocks.valueAtKey(Chunk.coord3DTo1D(x, y, z));
    }

    public Blocks[] treeTo1DArray() {
        // System.out.println(Config.CHUNK_3DCOORD_MAX_INDEX);
        Blocks[] result = new Blocks[Config.CHUNK_3DCOORD_MAX_INDEX + 1];
        Blocks v;
        Integer oldK = null;
        for(Integer k : blocks.treemap.keySet()){
            if(oldK != null){
                v = blocks.valueAtKey(oldK);
                //System.out.println("From " +oldK+ " to " + k + ": " + v);
                for(int i = oldK; i < k; i++) result[i] = v;
            }
            oldK = k;
        }
        v = blocks.valueAtKey(oldK);
        for(int i = oldK; i < result.length; i++) result[i] = v;
        return result;
    }

    public void treeFrom1DArray(Blocks[] array) {
        this.blocks = new IntervalMap<>();

        int start = 0;
        Blocks prev = array[0];

        for (int i = 0; i < array.length; i++) {
            if (array[i] != prev) {
                this.setBlocks(prev == null ? Blocks.AIR : prev, start, i);
                start = i;
            }

            prev = array[i];
        }

        this.setBlocks(prev, start, array.length);
    }

}
