package com.emamaker.voxeltest.intervaltrees.world.chunk;

import com.emamaker.voxeltest.intervaltrees.utils.Config;
import com.emamaker.voxeltest.intervaltrees.utils.OpenSimplexNoise;
import com.emamaker.voxeltest.intervaltrees.utils.SpaceFillingCurve;
import com.emamaker.voxeltest.intervaltrees.world.blocks.Blocks;

import java.util.Random;

public class ChunkGenerator {

    public void generateCube(Chunk c, Blocks block) {
        c.setBlocks(block, 0,
                Config.CHUNK_3DCOORD_MAX_INDEX);
    }

    public void arrayGenerateCorner(Chunk c) {
        Blocks[] array = new Blocks[Config.CHUNK_3DCOORD_MAX_INDEX + 1];

        int dist, hilbert;

        for (int i = 0; i < Config.CHUNK_SIZE; i++) {
            for (int j = 0; j < Config.CHUNK_SIZE; j++) {
                for (int k = 0; k < Config.CHUNK_SIZE; k++) {
                    // Distance from 0,0,0
                    dist = (int) Math.sqrt((i * i) + (j * j) + (k * k));
                    hilbert = SpaceFillingCurve.HILBERT_XYZ_ENCODE[i][j][k];
                    if (dist <= (int) (Config.CHUNK_SIZE / 3))
                        array[hilbert] = Blocks.STONE;
                    else if (dist > (int) (Config.CHUNK_SIZE / 3) && dist <= (int) (2 * Config.CHUNK_SIZE / 3))
                        array[hilbert] = Blocks.DIRT;
                    else if (dist > (int) (2 * Config.CHUNK_SIZE / 3) && dist <= (int) (Config.CHUNK_SIZE))
                        array[hilbert] = Blocks.GRASS;
                    else
                        array[hilbert] = Blocks.AIR;

                }
            }
        }

        c.treeFrom1DArray(array);
    }

    int GRASS_OFFSET = 80;
    int NOISE_GRASS_MULT = 20;
    int NOISE_DIRT_MULT = 3;
    int NOISE_DIRT_MIN = 2;
    float NOISE_DIRT_X_MULT = 0.001f;
    float NOISE_DIRT_Z_MULT = 0.001f;
    float NOISE_GRASS_X_MULT = 0.01f;
    float NOISE_GRASS_Z_MULT = 0.01f;

    OpenSimplexNoise noiseGen1 = new OpenSimplexNoise(new Random().nextInt());
    OpenSimplexNoise noiseGen2 = new OpenSimplexNoise(new Random().nextInt());
    // Generates smooth terrain using 2 layers of noise. This method takes advantage of how the interval map works and doesn't need a preliminary step to store data in an array
    // TODO (?): This can be optimizied for speed by only calculating noise levels at x-z coordinates once, and storing them in a 2d array

    // No LUT: about 135 ms
    // LUT: about 35 ms
    int[] grassNoiseLUT = new int[Config.CHUNK_SIZE*Config.CHUNK_SIZE];
    int[] dirtNoiseLUT = new int[Config.CHUNK_SIZE*Config.CHUNK_SIZE];
    public void generateNoise(Chunk c) {
        short coords[];
        int x, y, z;
        int grassNoise, dirtNoise, stoneLevel;
        int start = 0;
        Blocks prevBlock = null, block = Blocks.AIR;

        c.setState(Chunk.CHUNK_STATE_EMPTY, true);
        for(int i = 0; i < Config.CHUNK_SIZE*Config.CHUNK_SIZE; i++){
            grassNoiseLUT[i] = -1;
            dirtNoiseLUT[i] = -1;
        }
        for(int i = 0; i < Config.CHUNK_3DCOORD_MAX_INDEX + 1; i++){
                    coords = SpaceFillingCurve.HILBERT_XYZ_DECODE[i];
                    x = c.posx * Config.CHUNK_SIZE + coords[0];
                    y = c.posy * Config.CHUNK_SIZE + coords[1];
                    z = c.posz * Config.CHUNK_SIZE + coords[2];

                    if(grassNoiseLUT[coords[0] + Config.CHUNK_SIZE*coords[2]] == -1) grassNoiseLUT[coords[0] +Config.CHUNK_SIZE*coords[2]] = GRASS_OFFSET + (int) ((0.5d + noiseGen1.eval(x * NOISE_GRASS_X_MULT, z * NOISE_GRASS_Z_MULT)) * NOISE_GRASS_MULT);
                    if(dirtNoiseLUT[coords[0] + Config.CHUNK_SIZE*coords[2]] == -1) dirtNoiseLUT[coords[0] +Config.CHUNK_SIZE*coords[2]] =  NOISE_DIRT_MIN + (int) ((0.5f + noiseGen2.eval(x * NOISE_DIRT_X_MULT, z * NOISE_DIRT_Z_MULT)) * NOISE_DIRT_MULT);

                    grassNoise = grassNoiseLUT[coords[0] + Config.CHUNK_SIZE*coords[2]];
                    dirtNoise = dirtNoiseLUT[coords[0] + Config.CHUNK_SIZE*coords[2]];
                    stoneLevel = grassNoise - dirtNoise;

                    if (y < stoneLevel) block = Blocks.STONE;
                    else if (y >= stoneLevel && y < grassNoise) block = Blocks.DIRT;
                    else if (y == grassNoise) block = Blocks.GRASS;
                    else block = Blocks.AIR;

                    if(prevBlock == null) prevBlock = block;

                    if(prevBlock != block){
                        if(prevBlock != Blocks.AIR) c.setState(Chunk.CHUNK_STATE_EMPTY, false);
                        c.setBlocks(prevBlock, start, i);
                        start = i;
                    }
                    prevBlock = block;
        }

        if(prevBlock!= Blocks.AIR) c.setState(Chunk.CHUNK_STATE_EMPTY, false);
        c.setBlocks(prevBlock, start, Config.CHUNK_3DCOORD_MAX_INDEX + 1);
    }

    /*public void generateNoise(Chunk c) {
        short coords[];
        int x, y, z;
        int grassNoise, dirtNoise, stoneLevel;
        int start = 0;
        Blocks prevBlock = null, block = Blocks.AIR;
        int hilbert;
        Blocks[] array = new Blocks[Config.CHUNK_3DCOORD_MAX_INDEX + 1];
        c.setState(Chunk.CHUNK_STATE_EMPTY, true);
        //for(int i = 0; i < Config.CHUNK_3DCOORD_MAX_INDEX + 1; i++){
        for(int i = 0; i < Config.CHUNK_SIZE; i++ )
            for(int j = 0; j < Config.CHUNK_SIZE; j++ )
                for(int k = 0; k < Config.CHUNK_SIZE; k++ ){
                    //coords = SpaceFillingCurve.HILBERT_XYZ_DECODE[i];
                    hilbert = SpaceFillingCurve.HILBERT_XYZ_ENCODE[i][j][k];
                    x = c.posx*Config.CHUNK_SIZE + i;
                    y = c.posy*Config.CHUNK_SIZE + j;
                    z = c.posz*Config.CHUNK_SIZE + k;

                    grassNoise = GRASS_OFFSET + (int) ( (0.5d + noiseGen1.eval(x*NOISE_GRASS_X_MULT, z *NOISE_GRASS_Z_MULT) )*NOISE_GRASS_MULT);
                    dirtNoise = NOISE_DIRT_MIN +  (int)((0.5f + noiseGen2.eval(x*NOISE_DIRT_X_MULT, z*NOISE_DIRT_Z_MULT))*NOISE_DIRT_MULT);
                    stoneLevel = grassNoise - dirtNoise;

                    if(y < stoneLevel) block = Blocks.STONE;
                    else if(y >= stoneLevel && y < grassNoise) block = Blocks.DIRT;
                    else if(y == grassNoise) block = Blocks.GRASS;
                    else block = Blocks.AIR;

                    if(block != Blocks.AIR) c.setState(Chunk.CHUNK_STATE_EMPTY, false);
                    array[hilbert] = block;
                    //if(prevBlock == null) prevBlock = block;

                }
    }*/
}
