package com.emamaker.voxeltest.intervaltrees.world.chunk;

import com.emamaker.voxeltest.intervaltrees.utils.Config;
import com.emamaker.voxeltest.intervaltrees.utils.SpaceFillingCurve;
import com.emamaker.voxeltest.intervaltrees.world.WorldManager;
import com.emamaker.voxeltest.intervaltrees.world.blocks.Blocks;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Mesh;
import com.jme3.scene.VertexBuffer.Type;
import com.jme3.scene.shape.Box;
import com.jme3.util.BufferUtils;

import java.util.Arrays;

public class ChunkMesher {

    Blocks[] blocks;
    public void stupidArrayMeshing(WorldManager mgr, Chunk chunk) {
        blocks = chunk.treeTo1DArray();

        int idx;
        for (int i = 0; i < Config.CHUNK_SIZE; i++) {
            for (int j = 0; j < Config.CHUNK_SIZE; j++) {

                for (int k = 0; k < Config.CHUNK_SIZE; k++) {
                    idx = Chunk.coord3DTo1D(i, j, k);
                    if (blocks[idx] != Blocks.AIR) {
                        Box b = new Box(0.5f, 0.5f, 0.5f);

                        Geometry geom = new Geometry("Box", b);
                        Material mat = new Material(mgr.game.getAssetManager(), "Common/MatDefs/Misc/Unshaded.j3md");
                        // mat.getAdditionalRenderState().setWireframe(true);
                        geom.setMaterial(mat);
                        geom.getLocalTransform().setTranslation(i + 0.5f, j + 0.5f, k + 0.5f);

                        chunk.chunkNode.attachChild(geom);

                        if (blocks[idx] == Blocks.DIRT) {
                            mat.setColor("Color", ColorRGBA.Brown);
                        } else if (blocks[idx] == Blocks.STONE) {
                            mat.setColor("Color", ColorRGBA.Gray);
                        } else if (blocks[idx] == Blocks.GRASS) {
                            mat.setColor("Color", ColorRGBA.Green);
                        }
                        // else {
                        // mat.setColor("Color", ColorRGBA.Red);
                        // }
                    }
                }
            }
        }
    }

    /*
     * Taking inspiration from 0fps and the jme3 porting at
     * https://github.com/roboleary/GreedyMesh/blob/master/src/mygame/Main.java
     * 
     * By carefully re-reading the code and the blog post, I've come to the
     * realization that I wrote something very similar yet a lot messier (and
     * uglier) back in
     * https://github.com/EmaMaker/voxel-engine-jme3/blob/master/src/voxelengine
     * /world/Chunk.java
     * 
     * Reading roboleary's impl. I've learned how to optimize having to loop
     * across different planes everytime I change dimension without having to
     * write 3 separate 3-nested-for-loops
     */
    /*
     * It's not feasible to just create a new mesh everytime a quad needs placing.
     * This ends up creating TONS of meshes and the game will just lag out.
     * As I did in the past, it's better to create a single mesh for each chunk,
     * containing all the quads. In the future, maybe translucent blocks and liquids
     * will need a separate mesh, but still on a per-chunk basis
     */
    Vector3f[] vertices = new Vector3f[(Config.CHUNK_3DCOORD_MAX_INDEX + 1) * 8];
    float[] colors = new float[(Config.CHUNK_3DCOORD_MAX_INDEX + 1) * 8];
    short[] indices = new short[(Config.CHUNK_3DCOORD_MAX_INDEX + 1) * 8];

    short vIndex = 0;
    short iIndex = 0;
    short cIndex = 0;

    public void greedyMeshing(WorldManager mgr, Chunk chunk) {
        vIndex = 0;
        iIndex = 0;
        cIndex = 0;

        // convert tree to array since it is easier to work with it
        blocks = chunk.treeTo1DArray();

        if (blocks.length == 0)
            return;

        int k, l, u, v, w, h, n, j, i;
        int[] x = { 0, 0, 0 };
        int[] q = { 0, 0, 0 };
        final int[] du = new int[] { 0, 0, 0 };
        final int[] dv = new int[] { 0, 0, 0 };

        Blocks[] mask = new Blocks[Config.CHUNK_SIZE * Config.CHUNK_SIZE];
        // boolean backFace = false;
        for (boolean backFace = true, b = false; b != backFace; backFace = backFace && b, b = !b) {
            // iterate over 3 dimensions
            for (int dim = 0; dim < 3; dim++) {
                // offsets of other 2 axes
                u = (dim + 1) % 3;
                v = (dim + 2) % 3;

                x[0] = 0;
                x[1] = 0;
                x[2] = 0;

                q[0] = 0;
                q[1] = 0;
                q[2] = 0;
                q[dim] = 1; // easily mark which dimension we are comparing
                            // voxels
                            // on

                for (x[dim] = -1; x[dim] < Config.CHUNK_SIZE;) {
                    n = 0;

                    for (x[v] = 0; x[v] < Config.CHUNK_SIZE; x[v]++) {
                        for (x[u] = 0; x[u] < Config.CHUNK_SIZE; x[u]++) {
                            Blocks b1 = (x[dim] >= 0) ? blocks[SpaceFillingCurve.HILBERT_XYZ_ENCODE[x[0]][x[1]][x[2]]] : null;
                            Blocks b2 = (x[dim] < Config.CHUNK_SIZE - 1)
                                    ? blocks[SpaceFillingCurve.HILBERT_XYZ_ENCODE[x[0] + q[0]][x[1] + q[1]][x[2]+q[2]]]
                                    : null;

                            // This is the original line taken from rob's code, readapted (replace voxelFace
                            // with b1 and b2).
                            // mask[n++] = ((voxelFace != null && voxelFace1 != null &&
                            // voxelFace.equals(voxelFace1))) ? null : backFace ? voxelFace1 : voxelFace;

                            // Additionally checking whether b1 and b2 are AIR or null allows face culling,
                            // thus not rendering faces that cannot be seen
                            // Removing the control for null disables chunk borders
                            // This can be surely refactored in something that isn't such a big one-liner
                            mask[n++] = b1 != null && b2 != null && b1.equals(b2) ? null
                                    : backFace ? b1 == Blocks.AIR || b1 == null ? b2 : null
                                            : b2 == Blocks.AIR || b2 == null ? b1 : null;
                        }
                    }

                    x[dim]++;
                    n = 0;
                    // Actually generate the mesh from the mask. This is the
                    // same
                    // thing I used in my old crappy voxel engine
                    for (j = 0; j < Config.CHUNK_SIZE; j++) {
                        for (i = 0; i < Config.CHUNK_SIZE;) {

                            if (mask[n] != null) {
                                // First compute the width
                                for (w = 1; i + w < Config.CHUNK_SIZE && mask[n + w] != null
                                        && mask[n].equals(mask[n + w]); w++) {
                                }

                                boolean done = false;
                                for (h = 1; j + h < Config.CHUNK_SIZE; h++) {
                                    for (k = 0; k < w; k++) {
                                        if (mask[n + k + h * Config.CHUNK_SIZE] == null
                                                || !mask[n + k + h * Config.CHUNK_SIZE].equals(mask[n])) {
                                            done = true;
                                            break;
                                        }
                                    }
                                    if (done)
                                        break;
                                }

                                if (mask[n] != Blocks.AIR) {
                                    x[u] = i;
                                    x[v] = j;

                                    du[0] = 0;
                                    du[1] = 0;
                                    du[2] = 0;
                                    du[u] = w;

                                    dv[0] = 0;
                                    dv[1] = 0;
                                    dv[2] = 0;
                                    dv[v] = h;

                                    quad(new Vector3f(x[0], x[1], x[2]),
                                            new Vector3f(x[0] + du[0], x[1] + du[1], x[2] + du[2]),
                                            new Vector3f(x[0] + du[0] + dv[0], x[1] + du[1] + dv[1],
                                                    x[2] + du[2] + dv[2]),
                                            new Vector3f(x[0] + dv[0], x[1] + dv[1], x[2] + dv[2]),
                                            mask[n], backFace);
                                    // System.out.println(Arrays.toString(chunk.chunkNode.getChildren().toArray()));

                                }

                                for (l = 0; l < h; ++l) {
                                    for (k = 0; k < w; ++k) {
                                        mask[n + k + l * Config.CHUNK_SIZE] = null;
                                    }
                                }

                                /*
                                 * And then finally increment the counters and
                                 * continue
                                 */
                                i += w;
                                n += w;
                            } else {
                                i++;
                                n++;
                            }
                        }
                    }
                }
            }
        }

        if (vIndex > 0) {
            // Now, build a single mesh
            // System.out.println(Arrays.toString(Arrays.copyOf(vertices, vIndex )));
            // System.out.println(Arrays.toString(Arrays.copyOf(indices, iIndex )));

            Mesh mesh = new Mesh();
            mesh.setBuffer(Type.Position, 3,
                    BufferUtils.createFloatBuffer(Arrays.copyOf(vertices, vIndex)));
            mesh.setBuffer(Type.Color, 4, BufferUtils.createFloatBuffer(Arrays.copyOf(colors, cIndex)));
            mesh.setBuffer(Type.Index, 3, BufferUtils.createShortBuffer(Arrays.copyOf(indices, iIndex)));
            mesh.updateBound();

            Geometry geo = new Geometry("ColoredMesh", mesh);
            Material mat = new Material(mgr.game.getAssetManager(), "Common/MatDefs/Misc/Unshaded.j3md");

            mat.setBoolean("VertexColor", true);

             //mat.getAdditionalRenderState().setWireframe(true);
            // mat.getAdditionalRenderState().setFaceCullMode(FaceCullMode.Front);

            geo.setMaterial(mat);
            chunk.chunkNode.attachChild(geo);
        }
    }

    final int VOXEL_SIZE = 1;

    void quad(final Vector3f bottomLeft, final Vector3f topLeft, final Vector3f topRight,
            final Vector3f bottomRight, final Blocks block, boolean backFace) {

        // System.out.println(vIndex + ", " + iIndex);

        vertices[vIndex] = bottomLeft.mult(Config.VOXEL_SIZE);
        vertices[vIndex + 1] = bottomRight.mult(Config.VOXEL_SIZE);
        vertices[vIndex + 2] = topLeft.mult(Config.VOXEL_SIZE);
        vertices[vIndex + 3] = topRight.mult(Config.VOXEL_SIZE);

        if (backFace) {
            indices[iIndex] = (short) (vIndex + 2);
            indices[iIndex + 1] = (short) (vIndex + 0);
            indices[iIndex + 2] = (short) (vIndex + 1);
            indices[iIndex + 3] = (short) (vIndex + 1);
            indices[iIndex + 4] = (short) (vIndex + 3);
            indices[iIndex + 5] = (short) (vIndex + 2);
        } else {
            indices[iIndex] = (short) (vIndex + 2);
            indices[iIndex + 1] = (short) (vIndex + 3);
            indices[iIndex + 2] = (short) (vIndex + 1);
            indices[iIndex + 3] = (short) (vIndex + 1);
            indices[iIndex + 4] = (short) (vIndex + 0);
            indices[iIndex + 5] = (short) (vIndex + 2);
        }

        for (int i = cIndex; i < cIndex + 16; i += 4) {
            if (block == Blocks.DIRT) {

                colors[i] = 1.0f;
                colors[i + 1] = 0.0f;
                colors[i + 2] = 0.0f;
                colors[i + 3] = 1.0f;

            } else if (block == Blocks.GRASS) {

                colors[i] = 0.0f;
                colors[i + 1] = 1.0f;
                colors[i + 2] = 0.0f;
                colors[i + 3] = 1.0f;

            } else {
                colors[i] = 0.0f;
                colors[i + 1] = 0.0f;
                colors[i + 2] = 1.0f;
                colors[i + 3] = 1.0f;
            }
        }

        // next time it will be the index of the first vertex that was added
        vIndex += 4;
        iIndex += 6;
        cIndex += 16;
    }
}